from typing import Any, Dict, List, Optional, Union

import itertools
import sys
from abc import ABC, abstractmethod
from datetime import datetime
from pathlib import Path

import torch
from loguru import logger
from pytorch_lightning import Callback, Trainer

from lighter import LighterSystem
from lighter.callbacks.utils import gather_tensors, parse_data


class LighterBaseWriter(ABC, Callback):
    """Base class for a Writer. Override `self.write()` to define how a prediction should be saved.
    `LighterBaseWriter` sets up the write directory, and defines `on_predict_batch_end` and
    `on_predict_epoch_end`. `write_interval` specifies which of the two should the writer call.

    Args:
        write_dir (str): the Writer will create a directory inside of `write_dir` with date
            and time as its name and store the predictions there.
        write_format (Optional[Union[str, List[str], Dict[str, str], Dict[str, List[str]]]]):
            type in which the predictions will be stored. Passed automatically to the `write()`
            abstract method and can be used to support writing different types. Should the Writer
            support only one type, this argument can be removed from the overriden `__init__()`'s
            arguments and set `self.write_format = None`.
        write_interval (str, optional): whether to write on each step or at the end of the prediction epoch.
            Defaults to "step".
    """

    def __init__(
        self,
        write_dir: str,
        write_format: Optional[Union[str, List[str], Dict[str, str], Dict[str, List[str]]]],
        write_interval: str = "step",
    ) -> None:
        self.write_dir = Path(write_dir) / datetime.now().strftime("%Y%m%d_%H%M%S")
        self.write_format = write_format
        self.write_interval = write_interval

        self.parsed_write_format = None

    @abstractmethod
    def write(
        self,
        idx: int,
        identifier: Optional[str],
        tensor: torch.Tensor,
        write_format: Optional[Union[str, List[str], Dict[str, str], Dict[str, List[str]]]],
    ):
        """This method must be overridden to specify how a tensor should be saved. If the Writer
        supports multiple types of saving, handle the `write_format` argument with an if-else statement.

        If the Writer only supports one type, remove `write_format` from the overridden
        `__init__()` method and set `self.write_format=None`.

        The `idx` and `identifier` arguments can be used to specify the name of the file
        or the row and column of a table for the prediction.

        Parameters:
            idx (int): The index of the prediction.
            identifier (Optional[str]): The identifier of the prediction. It will be `None` if there's
                only one prediction, an index if the prediction is a list of predictions, a key if it's
                a dict of predictions, and a key_index if it's a dict of list of predictions.
            tensor (torch.Tensor): The predicted tensor.
            write_format (Optional[Union[str, List[str], Dict[str, str], Dict[str, List[str]]]]):
                Specifies how to write the predictions. If it's a single string value, the predictions
                will be saved under that type regardless of whether they are single- or multi-output
                predictions. To write different outputs in the multi-output predictions using different
                methods, use the appropriate format for `write_format`.
        """

    def setup(self, trainer: Trainer, pl_module: LighterSystem, stage: str) -> None:
        if stage != "predict":
            return

        if self.write_interval not in ["step", "epoch"]:
            logger.error("`write_interval` must be either 'step' or 'epoch'.")
            sys.exit()

        # Broadcast the `write_dir` so that all ranks write their predictions there.
        self.write_dir = trainer.strategy.broadcast(self.write_dir)
        # Let rank 0 create the `write_dir`.
        if trainer.is_global_zero:
            self.write_dir.mkdir(parents=True)
        # If `write_dir` does not exist, the ranks are not on the same storage.
        if not self.write_dir.exists():
            logger.error(
                f"Rank {trainer.global_rank} is not on the same storage as rank 0."
                "Please run the prediction only on nodes that are on the same storage."
            )
            sys.exit()

    def on_predict_batch_end(
        self, trainer: Trainer, pl_module: LighterSystem, outputs: Any, batch: Any, batch_idx: int, dataloader_idx: int
    ) -> None:
        if self.write_interval != "step":
            return
        indices = trainer.predict_loop.epoch_loop.current_batch_indices
        self._on_batch_or_epoch_end(outputs, indices)

    def on_predict_epoch_end(self, trainer: Trainer, pl_module: LighterSystem, outputs: List[Any]) -> None:
        if self.write_interval != "epoch":
            return
        # Only one epoch when predicting, index the lists of outputs and batch indices accordingly.
        indices = trainer.predict_loop.epoch_batch_indices[0]
        outputs = outputs[0]
        # Flatten so that each output sample corresponds to its index.
        indices = list(itertools.chain(*indices))
        # Remove the batch dimension since every output is a single sample.
        outputs = [output.squeeze(0) for output in outputs]
        # Gather the output tensors from all samples into a single structure rather than having one structures for each sample.
        outputs = gather_tensors(outputs)
        self._on_batch_or_epoch_end(outputs, indices)

    def _on_batch_or_epoch_end(self, outputs, indices):
        """Iterate through each output and save it in the specified format. The outputs and indices are automatically
        split individually by PyTorch Lightning."""
        # Sanity check. Should never happen. If it does, something is wrong with the Trainer.
        assert len(indices) == len(outputs)
        # `idx` is the index of the input sample, `output` is the output of the model for that sample.
        for idx, output in zip(indices, outputs):
            # Parse the outputs into a structure ready for writing.
            parsed_output = parse_data(output)
            # Parse `self.write_format`. If multi-value, check if its structure matches `parsed_output`'s structure.
            if self.parsed_write_format is None:
                self.parsed_write_format = self._parse_write_format(self.write_format, parsed_output)
            # Iterate through each prediction for the `idx`-th input sample.
            for identifier, tensor in parsed_output.items():
                # Save the prediction in the specified format.
                self.write(idx, identifier, tensor.detach().cpu(), self.parsed_write_format[identifier])

    def _parse_write_format(self, write_format, parsed_outputs: Dict[str, Any]):
        # If `write_format` is a string (single value), all outputs will be saved in that specified format.
        if isinstance(write_format, str):
            parsed_write_format = {key: write_format for key in parsed_outputs}
        # Otherwise, `write_format` needs to match the structure of the outputs in order to assign each tensor its type.
        else:
            parsed_write_format = parse_data(write_format)
            if not set(parsed_write_format) == set(parsed_outputs):
                logger.error("`write_format` structure does not match the prediction's structure.")
                sys.exit()
        return parsed_write_format



from typing import Any, Callable, List

import inspect
import sys

from loguru import logger


def ensure_list(vals: Any) -> List:
    """Wrap the input into a list if it is not a list. If it is a None, return an empty list.

    Args:
        vals (Any): input to wrap into a list.

    Returns:
        List: output list.
    """
    if isinstance(vals, list):
        return vals
    if isinstance(vals, tuple):
        return list(vals)
    if vals is None:
        return []
    return [vals]


def setattr_dot_notation(obj: Callable, attr: str, value: Any):
    """Set object's attribute. Supports dot notation.

    Args:
        obj (Callable): object.
        attr (str): attribute name of the object.
        value (Any): value to set the attribute to.
    """
    if "." not in attr:
        if not hasattr(obj, attr):
            logger.info(f"`{get_name(obj, True)}` has no attribute `{attr}`. Exiting.")
            sys.exit()
        setattr(obj, attr, value)
    # Solve recursively if the attribute is defined in dot-notation
    else:
        obj_name, attr = attr.split(".", maxsplit=1)
        setattr_dot_notation(getattr(obj, obj_name), attr, value)


def hasarg(_callable: Callable, arg_name: str) -> bool:
    """Check if a function, class, or method has an argument with the specified name.

    Args:
        _callable (Callable): function, class, or method to inspect.
        arg_name (str): argument name to check for.

    Returns:
        bool: `True` if the argument if the specified name exists.
    """

    args = inspect.signature(_callable).parameters.keys()
    return arg_name in args


def countargs(_callable: Callable) -> bool:
    """Count the number of arguments that a function, class, or method accepts.

    Args:
        callable (Callable): function, class, or method to inspect.

    Returns:
        int: number of arguments that it accepts.
    """
    return len(inspect.signature(_callable).parameters.keys())


def get_name(_callable: Callable, include_module_name: bool = False) -> str:
    """Get the name of an object, class or function.

    Args:
        _callable (Callable): object, class or function.
        include_module_name (bool, optional): whether to include the name of the module from
            which it comes. Defaults to False.

    Returns:
        str: name
    """
    name = type(_callable).__name__ if isinstance(_callable, object) else _callable.__name__
    if include_module_name:
        module = type(_callable).__module__ if isinstance(_callable, object) else _callable.__module__
        name = f"{module}.{name}"
    return name



from typing import List, Optional, Union

from loguru import logger
from torch.nn import Module

from lighter.utils.misc import ensure_list


class LighterFreezer:
    """
    Freezes the parameters/layers of a model. Can be run indefinitely or until a specified step or epoch.
    `names` and`name_starts_with` can be used to specify which parameters to freeze.
    If both are specified, the parameters that match any of the two will be frozen.

    Args:
        names (str, List[str], optional): The names of the parameters to be frozen. Defaults to None.
        name_starts_with (str, List[str], optional): The prefixes of the parameter names to be frozen. Defaults to None.
        until_step (int, optional): The maximum step to freeze parameters until. Defaults to None.
        until_epoch (int, optional): The maximum epoch to freeze parameters until. Defaults to None.

    Raises:
        ValueError: If neither `names` nor `name_starts_with` are specified.
        ValueError: If both `until_step` and `until_epoch` are specified.

    """

    def __init__(
        self,
        names: Optional[Union[str, List[str]]] = None,
        name_starts_with: Optional[Union[str, List[str]]] = None,
        until_step: int = None,
        until_epoch: int = None,
    ) -> None:
        if names is None and name_starts_with is None:
            raise ValueError("At least one of `names` or `name_starts_with` must be specified.")

        if until_step is not None and until_epoch is not None:
            raise ValueError("Only one of `until_step` or `until_epoch` can be specified.")

        self.names = ensure_list(names)
        self.name_starts_with = ensure_list(name_starts_with)
        self.until_step = until_step
        self.until_epoch = until_epoch

        self._frozen_layers_reported = False

    def __call__(self, model: Module, current_step: int, current_epoch: int) -> None:
        """
        Freezes the parameters of the model.

        Args:
            model (Module): The PyTorch model whose parameters need to be frozen.
            current_step (int): The current step.
            current_epoch (int): The current epoch.

        """
        if self.until_step is not None and current_step > self.until_step:
            return
        if self.until_epoch is not None and current_epoch > self.until_epoch:
            return

        # Keep track of the frozen layers to log them on the first call.
        frozen_layers = []

        # Freeze the specified parameters.
        for name, param in model.named_parameters():
            if self.names and name in self.names:
                param.requires_grad = False
                frozen_layers.append(name)
            elif self.name_starts_with and any(name.startswith(prefix) for prefix in self.name_starts_with):
                param.requires_grad = False
                frozen_layers.append(name)
            else:
                param.requires_grad = True

        # Log the frozen layers' names, only on the first call.
        if not self._frozen_layers_reported:
            self._frozen_layers_reported = True
            logger.info(
                "Freezing the following layers"
                + (f" until step {self.until_step}" if self.until_step is not None else "")
                + (f" until epoch {self.until_epoch}" if self.until_epoch is not None else "")
                + f": {frozen_layers}"
            )



from typing import Any

import importlib
import sys
from pathlib import Path

from loguru import logger

OPTIONAL_IMPORTS = {}


def import_module_from_path(module_name: str, module_path: str) -> None:
    """Given the path to a module, import it, and name it as specified.

    Args:
        module_name (str): what to name the imported module.
        module_path (str): path to the module to load.
    """
    # Based on https://stackoverflow.com/a/41595552.
    module_path = Path(module_path).resolve() / "__init__.py"
    print(module_path)
    if not module_path.is_file():
        logger.error(f"No `__init__.py` in `{module_path}`. Exiting.")
        sys.exit()
    spec = importlib.util.spec_from_file_location(module_name, str(module_path))
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    sys.modules[module_name] = module
    logger.info(f"{module_path.parent} imported as '{module_name}' module.")


def import_attr(module_attr: str) -> Any:
    """Import using dot-notation string, e.g., 'torch.nn.Module'.

    Args:
        module_attr (str): dot-notation path to the attribute.

    Returns:
        Any: imported attribute.
    """
    # Split module from attribute name
    module, attr = module_attr.rsplit(".", 1)
    # Import the module
    module = __import__(module, fromlist=[attr])
    # Get the attribute from the module
    return getattr(module, attr)



import sys

import torch
import torchvision
from loguru import logger
from monai.transforms import DivisiblePad
from monai.utils.module import optional_import

from lighter.callbacks.utils import preprocess_image
from lighter.callbacks.writer.base import LighterBaseWriter
from lighter.utils.dynamic_imports import OPTIONAL_IMPORTS


class LighterFileWriter(LighterBaseWriter):
    def write(self, idx, identifier, tensor, write_format):
        filename = f"{write_format}" if identifier is None else f"{identifier}_{write_format}"
        write_dir = self.write_dir / str(idx)
        write_dir.mkdir()

        if write_format is None:
            pass

        # Tensor
        elif write_format == "tensor":
            path = write_dir / f"{filename}.pt"
            torch.save(tensor, path)

        # Image
        elif write_format == "image":
            path = write_dir / f"{filename}.png"
            torchvision.io.write_png(preprocess_image(tensor), path)

        # Video
        elif write_format == "video":
            path = write_dir / f"{filename}.mp4"
            # Video tensor must be divisible by 2. Pad the height and width.
            tensor = DivisiblePad(k=(0, 2, 2), mode="minimum")(tensor)
            # Video tensor must be THWC. Permute CTHW -> THWC.
            tensor = tensor.permute(1, 2, 3, 0)
            # Video tensor must have 3 channels (RGB). Repeat the channel dim to convert grayscale to RGB.
            if tensor.shape[-1] == 1:
                tensor = tensor.repeat(1, 1, 1, 3)
            # Video tensor must be in the range [0, 1]. Scale to [0, 255].
            tensor = (tensor * 255).to(torch.uint8)
            torchvision.io.write_video(str(path), tensor, fps=24)

        # Scalar
        elif write_format == "scalar":
            raise NotImplementedError

        # Audio
        elif write_format == "audio":
            raise NotImplementedError

        else:
            logger.error(f"`write_format` '{write_format}' not supported.")
            sys.exit()


def write_sitk_image(path: str, tensor: torch.Tensor) -> None:
    """Write a SimpleITK image to disk.

    Args:
        path (str): path to write the image.
        tensor (torch.Tensor): tensor to write.
    """
    if "sitk" not in OPTIONAL_IMPORTS:
        OPTIONAL_IMPORTS["sitk"], sitk_available = optional_import("SimpleITK")
        if not sitk_available:
            raise ModuleNotFoundError("SimpleITK not installed. To install it, run `pip install SimpleITK`. Exiting.")
    sitk_image = OPTIONAL_IMPORTS["sitk"].GetImageFromArray(tensor.cpu().numpy())
    OPTIONAL_IMPORTS["sitk"].WriteImage(sitk_image, str(path), True)



from typing import Any, Dict

import sys
from functools import partial

import fire
import yaml
from loguru import logger
from monai.bundle.config_parser import ConfigParser
from pytorch_lightning import seed_everything

from lighter.utils.dynamic_imports import import_module_from_path
from lighter.utils.misc import ensure_list


def interface():
    """Defines the command line interface for running Trainer's methods. The available methods are:
    - fit
    - validate
    - test
    - predict
    - tune
    """
    commands = {}
    # All Trainer methods' names
    for name in ["fit", "validate", "test", "predict", "tune"]:
        # Creates command for fire to run. The command is the Trainer method's name. Fire passes
        # **kwargs to the function.
        commands[name] = partial(run_trainer_method, name)

    fire.Fire(commands)


def parse_config(**kwargs):
    """
    Parses configuration files and updates the provided parser
    with given keyword arguments. Returns an updated parser object.

    Args:
        **kwargs (dict): Keyword arguments containing configuration data.
            config_file (str): Path to the main configuration file.
            args_file (str, optional): Path to secondary configuration file for additional arguments.
            Additional key-value pairs can also be provided to be added or updated in the parser.

    Returns:
        ConfigParser: An instance of ConfigParser with parsed and merged configuration data.
    """
    parser = ConfigParser()
    parser.read_config(kwargs["config_file"])

    if "args_file" in kwargs:
        args = ConfigParser.load_config_file(kwargs["args_file"])
        parser.update(pairs=args)

    parser.update(pairs=kwargs)

    return parser


def run_trainer_method(method: Dict, **kwargs: Any):
    """Call monai.bundle.run() on a Trainer method. If a project path
    is defined in the config file(s), import it.

    Args:
        method_name: name of the Trainer method to run. ["fit", "validate", "test", "predict", "tune"].
        **kwargs (Any): keyword arguments passed to the `monai.bundle.run` function.
    """
    # Sets the random seed to `PL_GLOBAL_SEED` env variable. If not specified, it picks a random seed.
    seed_everything()

    # Check that a config file is specified.
    if "config_file" not in kwargs:
        raise ValueError("No config file specified. Exiting.")

    # Import the project as a module.
    project_imported = False
    # Handle multiple configs. Start from the config file specified last as it overrides the previous ones.
    for config in reversed(ensure_list(kwargs["config_file"])):
        with open(config, encoding="utf-8") as config:
            config = yaml.safe_load(config)
            if "project" not in config:
                continue
            # Only one config file can specify the project path
            if project_imported:
                logger.error("`project` must be specified in one config only. Exiting.")
                sys.exit()
            # Import it as a module named 'project'.
            import_module_from_path("project", config["project"])
            project_imported = True

    # Parse the config file(s).
    parser = parse_config(**kwargs)
    trainer = parser.get_parsed_content("trainer")
    system = parser.get_parsed_content("system")

    # Run the Trainer method.
    if not hasattr(trainer, method):
        raise ValueError(f"Trainer has no method named {method}.")

    getattr(trainer, method)(system)



from typing import Any, Dict, List, Union

import itertools
import sys

import pandas as pd
from loguru import logger
from pytorch_lightning import Trainer

from lighter import LighterSystem
from lighter.callbacks.writer.base import LighterBaseWriter


class LighterTableWriter(LighterBaseWriter):
    def __init__(self, write_dir: str, write_format: Union[str, List[str], Dict[str, str], Dict[str, List[str]]]) -> None:
        super().__init__(write_dir, write_format, write_interval="epoch")
        self.csv_records = {}

    def write(self, idx, identifier, tensor, write_format):
        # Column name will be set to 'pred' if the identifier is None.
        column = "pred" if identifier is None else identifier

        if write_format is None:
            record = None
        elif write_format == "tensor":
            record = tensor.tolist()
        elif write_format == "scalar":
            raise NotImplementedError
        else:
            logger.error(f"`write_format` '{write_format}' not supported.")
            sys.exit()

        if idx not in self.csv_records:
            self.csv_records[idx] = {column: record}
        else:
            self.csv_records[idx][column] = record

    def on_predict_epoch_end(self, trainer: Trainer, pl_module: LighterSystem, outputs: List[Any]) -> None:
        super().on_predict_epoch_end(trainer, pl_module, outputs)

        csv_path = self.write_dir / "predictions.csv"
        logger.info(f"Saving the predictions to {csv_path}")

        # Sort the dict of dicts by key and turn it into a list of dicts.
        self.csv_records = [self.csv_records[key] for key in sorted(self.csv_records)]
        # Gather the records from all ranks when in DDP.
        if trainer.world_size > 1:
            # Since `all_gather` supports tensors only, mimic the behavior using `broadcast`.
            ddp_csv_records = [self.csv_records] * trainer.world_size
            for rank in range(trainer.world_size):
                # Broadcast the records from the current rank and save it at its designated position.
                ddp_csv_records[rank] = trainer.strategy.broadcast(ddp_csv_records[rank], src=rank)
            # Combine the records from all ranks. List of lists of dicts -> list of dicts.
            self.csv_records = list(itertools.chain(*ddp_csv_records))

        # Create a dataframe and save it.
        pd.DataFrame(self.csv_records).to_csv(csv_path)



from typing import Any, Dict, List, Optional, Tuple, Union

import torch
import torchvision


def get_lighter_mode(lightning_stage: str) -> str:
    """Converts the name of a PyTorch Lightnig stage to the name of its corresponding Lighter mode.

    Args:
        lightning_stage (str): stage in which PyTorch Lightning Trainer is. Can be accessed using `trainer.state.stage`.

    Returns:
        str: name of the Lighter mode.
    """
    lightning_to_lighter = {"train": "train", "validate": "val", "test": "test"}
    return lightning_to_lighter[lightning_stage]


def is_data_type_supported(data: Any) -> bool:
    """Check the input data for its type. Valid data types are:
        - torch.Tensor
        - List[torch.Tensor]
        - Tuple[torch.Tensor]
        - Dict[str, torch.Tensor]
        - Dict[str, List[torch.Tensor]]
        - Dict[str, Tuple[torch.Tensor]]

    Args:
        data (Any): input data to check

    Returns:
        bool: True if the data type is supported, False otherwise.
    """
    if isinstance(data, dict):
        is_valid = all(is_data_type_supported(elem) for elem in data.values())
    elif isinstance(data, (list, tuple)):
        is_valid = all(is_data_type_supported(elem) for elem in data)
    elif isinstance(data, torch.Tensor):
        is_valid = True
    else:
        is_valid = False
    return is_valid


def parse_data(
    data: Union[Any, List[Any], Dict[str, Any], Dict[str, List[Any]], Dict[str, Tuple[Any]]]
) -> Dict[Optional[str], Any]:
    """Parse the input data as follows:
        - If dict, go over all keys and values, unpacking list and tuples, and assigning them all
          a unique identifier based on the original key and their position if they were a list/tuple.
        - If list/tuple, enumerate them and use their position as key for each value of the list/tuple.
        - If any other type, return it as-is with the key set to 'None'. A 'None' key indicates that no
          identifier is needed because no parsing ocurred.

    Args:
        data (Any, List[Any], Dict[str, Any], Dict[str, List[Any]], Dict[str, Tuple[Any]):
            input data to parse.

    Returns:
        Dict[Optional[str], Any]: a dict where key is either a string
            identifier or `None`, and value the parsed output.
    """
    result = {}
    if isinstance(data, dict):
        for key, value in data.items():
            if isinstance(value, (list, tuple)):
                for idx, element in enumerate(value):
                    result[f"{key}_{idx}" if len(value) > 1 else key] = element
            else:
                result[key] = value
    elif isinstance(data, (list, tuple)):
        for idx, element in enumerate(data):
            result[str(idx)] = element
    else:
        result[None] = data
    return result


def gather_tensors(
    inputs: Union[List[Union[torch.Tensor, List, Tuple, Dict]], Tuple[Union[torch.Tensor, List, Tuple, Dict]]]
) -> Union[List, Dict]:
    """Recursively gather tensors. Tensors can be standalone or inside of other data structures (list/tuple/dict).
    An input list of tensors is returned as-is. Given an input list of data structures with tensors, this function
    will gather all tensors into a list and save it under a single data structure. Assumes that all elements of
    the input list have the same type and structure.

    Args:
        inputs (List[Union[torch.Tensor, List, Tuple, Dict]], Tuple[Union[torch.Tensor, List, Tuple, Dict]]):
            They can be:
            - List/Tuples of Dictionaries, each containing tensors to be gathered by their key.
            - List/Tuples of Lists/tuples, each containing tensors to be gathered by their position.
            - List/Tuples of Tensors, returned as-is.
            - Nested versions of the above.
            The input data structure must be the same for all elements of the list. They can be arbitrarily nested.

    Returns:
        Union[List, Dict]: The gathered tensors.
    """
    # List of dicts.
    if isinstance(inputs[0], dict):
        keys = inputs[0].keys()
        return {key: gather_tensors([input[key] for input in inputs]) for key in keys}
    # List of lists or tuples.
    elif isinstance(inputs[0], (list, tuple)):
        return [gather_tensors([input[idx] for input in inputs]) for idx in range(len(inputs[0]))]
    # List of tensors.
    elif isinstance(inputs[0], torch.Tensor):
        return inputs
    else:
        raise TypeError(f"Type `{type(inputs[0])}` not supported.")


def preprocess_image(image: torch.Tensor) -> torch.Tensor:
    """Preprocess the image before logging it. If it is a batch of multiple images,
    it will create a grid image of them. In case of 3D, a single image is displayed
    with slices stacked vertically, while a batch of 3D images as a grid where each
    column is a different 3D image.
    Args:
        image (torch.Tensor): 2D or 3D image tensor.
    Returns:
        torch.Tensor: image ready for logging.
    """
    image = image.detach().cpu()
    # If 3D (BCDHW), concat the images vertically and horizontally.
    if image.ndim == 5:
        shape = image.shape
        # BCDHW -> BC(D*H)W. Combine slices of a 3D images vertically into a single 2D image.
        image = image.view(shape[0], shape[1], shape[2] * shape[3], shape[4])
        # BCDHW -> 1CDH(B*W). Concat images in the batch horizontally, and unsqueeze to add back the B dim.
        image = torch.cat([*image], dim=-1).unsqueeze(0)
    # If only one image in the batch, select it and return it. Same happens when the images are 3D as they
    # are combined into a single image. `make_grid` is called when a batch of multiple 2D image is provided.
    return image[0] if image.shape[0] == 1 else torchvision.utils.make_grid(image, nrow=8)



from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import sys
from functools import partial

import pytorch_lightning as pl
import torch
from loguru import logger
from torch.nn import Module
from torch.optim import Optimizer
from torch.utils.data import DataLoader, Dataset, Sampler
from torchmetrics import Metric, MetricCollection

from lighter.utils.collate import collate_fn_replace_corrupted
from lighter.utils.misc import countargs, ensure_list, get_name, hasarg
from lighter.utils.model import reshape_pred_if_single_value_prediction


class LighterSystem(pl.LightningModule):
    """_summary_

    Args:
        model (Module): the model.
        batch_size (int): batch size.
        drop_last_batch (bool, optional): whether the last batch in the dataloader
            should be dropped. Defaults to False.
        num_workers (int, optional): number of dataloader workers. Defaults to 0.
        pin_memory (bool, optional): whether to pin the dataloaders memory. Defaults to True.
        optimizers (Optional[Union[Optimizer, List[Optimizer]]], optional):
            a single or a list of optimizers. Defaults to None.
        schedulers (Optional[Union[Callable, List[Callable]]], optional):
            a single or a list of schedulers. Defaults to None.
        criterion (Optional[Callable], optional):
            criterion/loss function. Defaults to None.
        cast_target_dtype_to (Optional[str], optional): whether to cast the target to the
            specified type before calculating the loss. May be necessary for some criterions.
            Defaults to None.
        post_criterion_activation (Optional[str], optional): some criterions
            (e.g. BCEWithLogitsLoss) require non-activated prediction for their calculaiton.
            However, to calculate the metrics and log the data, it may be necessary to activate
            the predictions. Defaults to None.
        inferer (Optional[Callable], optional): the inferer must be a class with a `__call__`
            method that accepts two arguments - the input to infer over, and the model itself.
            Used in 'val', 'test', and 'predict' mode, but not in 'train'. Typically, an inferer
            is a sliding window or a patch-based inferer that will infer over the smaller parts of
            the input, combine them, and return a single output. The inferers provided by MONAI
            cover most of such cases (https://docs.monai.io/en/stable/inferers.html). Defaults to None.
        freezer (Optional[Callable], optional): the freezer must be a class with a `__call__`
            method that accepts three arguments - the model, the step, and the epoch number.
            Use `lighter.utils.freezer.LighterFreezer` or implement your own based on it. Defaults to None.
        train_metrics (Optional[Union[Metric, List[Metric]]], optional): training metric(s).
            They have to be implemented using `torchmetrics`. Defaults to None.
        val_metrics (Optional[Union[Metric, List[Metric]]], optional): validation metric(s).
            They have to be implemented using `torchmetrics`. Defaults to None.
        test_metrics (Optional[Union[Metric, List[Metric]]], optional): test metric(s).
            They have to be implemented using `torchmetrics`. Defaults to None.
        train_dataset (Optional[Union[Dataset, List[Dataset]]], optional): training dataset(s).
            Defaults to None.
        val_dataset (Optional[Union[Dataset, List[Dataset]]], optional): validation dataset(s).
            Defaults to None.
        test_dataset (Optional[Union[Dataset, List[Dataset]]], optional): test dataset(s).
            Defaults to None.
        predict_dataset (Optional[Union[Dataset, List[Dataset]]], optional): predict dataset(s).
            Defaults to None.
        train_sampler (Optional[Sampler], optional): training sampler(s). Defaults to None.
        val_sampler (Optional[Sampler], optional): validation sampler(s). Defaults to None.
        test_sampler (Optional[Sampler], optional):  test sampler(s). Defaults to None.
        predict_sampler (Optional[Sampler], optional):  predict sampler(s). Defaults to None.
        train_collate (Optional[Callable], optional): custom training collate function. Defaults to None.
        val_collate (Optional[Callable], optional): custom validation collate function. Defaults to None.
        test_collate (Optional[Callable], optional):  custom test collate function. Defaults to None.
        predict_collate (Optional[Callable], optional):  custom predict collate function. Defaults to None.
    """

    def __init__(
        self,
        model: Module,
        batch_size: int,
        drop_last_batch: bool = False,
        num_workers: int = 0,
        pin_memory: bool = True,
        optimizers: Optional[Union[Optimizer, List[Optimizer]]] = None,
        schedulers: Optional[Union[Callable, List[Callable]]] = None,
        criterion: Optional[Callable] = None,
        cast_target_dtype_to: Optional[str] = None,
        post_criterion_activation: Optional[str] = None,
        inferer: Optional[Callable] = None,
        freezer: Optional[Callable] = None,
        train_metrics: Optional[Union[Metric, List[Metric]]] = None,
        val_metrics: Optional[Union[Metric, List[Metric]]] = None,
        test_metrics: Optional[Union[Metric, List[Metric]]] = None,
        train_dataset: Optional[Union[Dataset, List[Dataset]]] = None,
        val_dataset: Optional[Union[Dataset, List[Dataset]]] = None,
        test_dataset: Optional[Union[Dataset, List[Dataset]]] = None,
        predict_dataset: Optional[Union[Dataset, List[Dataset]]] = None,
        train_sampler: Optional[Sampler] = None,
        val_sampler: Optional[Sampler] = None,
        test_sampler: Optional[Sampler] = None,
        predict_sampler: Optional[Sampler] = None,
        train_collate: Optional[Callable] = None,
        val_collate: Optional[Callable] = None,
        test_collate: Optional[Callable] = None,
        predict_collate: Optional[Callable] = None,
    ) -> None:
        super().__init__()
        # Bypass LightningModule's check for default methods. We define them in self.setup().
        self._init_placeholders_for_dataloader_and_step_methods()

        # Model setup
        self.model = model
        self.batch_size = batch_size
        self.drop_last_batch = drop_last_batch

        # Criterion, optimizer, and scheduler
        self.criterion = criterion
        self.optimizers = ensure_list(optimizers)
        self.schedulers = ensure_list(schedulers)

        # Datasets
        self.train_dataset = train_dataset
        self.val_dataset = val_dataset
        self.test_dataset = test_dataset
        self.predict_dataset = predict_dataset
        self.num_workers = num_workers
        self.pin_memory = pin_memory

        # Samplers
        self.train_sampler = train_sampler
        self.val_sampler = val_sampler
        self.test_sampler = test_sampler
        self.predict_sampler = predict_sampler

        # Collate functions
        self.train_collate = train_collate
        self.val_collate = val_collate
        self.test_collate = test_collate
        self.predict_collate = predict_collate

        # Metrics
        self.train_metrics = MetricCollection(ensure_list(train_metrics))
        self.val_metrics = MetricCollection(ensure_list(val_metrics))
        self.test_metrics = MetricCollection(ensure_list(test_metrics))

        # Criterion-specific activation function and data type casting
        self._post_criterion_activation = post_criterion_activation
        self._cast_target_dtype_to = cast_target_dtype_to

        # Inferer for val, test, and predict
        self.inferer = inferer

        # Layer freezer
        self.freezer = freezer

        # Checks
        self._lightning_module_methods_defined = False
        self._target_not_used_reported = False
        self._batch_type_reported = False

    def forward(self, input: Union[torch.Tensor, List[torch.Tensor], Tuple[torch.Tensor], Dict[str, torch.Tensor]]) -> Any:
        """Forward pass. Multi-input models are supported.

        Args:
            input (torch.Tensor, List[torch.Tensor], Tuple[torch.Tensor], Dict[str, torch.Tensor]): input to the model.

        Returns:
            Any: output of the model.
        """
        # Keyword arguments to pass to the forward method
        kwargs = {}
        if hasarg(self.model.forward, "epoch"):
            # Add `epoch` argument if forward accepts it
            kwargs["epoch"] = self.current_epoch
        if hasarg(self.model.forward, "step"):
            # Add `step` argument if forward accepts it
            kwargs["step"] = self.global_step

        # Type not supported.
        if not isinstance(input, (torch.Tensor, tuple, list, dict)):
            logger.error(f"Input type '{type(input)}' not supported.")
            sys.exit()

        # Freeze the layers if specified so.
        if self.freezer is not None:
            self.freezer(self.model, self.global_step, self.current_epoch)

        # Unpack Tuple or List. Only if num of args passed is less than or equal to num of args accepted.
        if isinstance(input, (tuple, list)) and (len(input) + len(kwargs)) <= countargs(self.model):
            return self.model(*input, **kwargs)
        # Unpack Dict. Only if dict's keys match criterion's keyword arguments.
        elif isinstance(input, dict) and all(hasarg(self.model, name) for name in input):
            return self.model(**input, **kwargs)
        # Tensor, Tuple, List, or Dict, as-is, not unpacked.
        else:
            return self.model(input, **kwargs)

    def _base_step(self, batch: Union[List, Tuple], batch_idx: int, mode: str) -> Union[Dict[str, Any], Any]:
        """Base step for all modes ("train", "val", "test", "predict")

        Args:
            batch (List, Tuple):
                output of the DataLoader and input to the model.
            batch_idx (int): index of the batch. Not used, but PyTorch Lightning requires it.
            mode (str): mode in which the system is.

        Returns:
            Union[Dict[str, Any], Any]: For the training, validation and test step, it returns
                a dict containing loss, metrics, input, target, and pred. Loss will be `None`
                for the test step. Metrics will be `None` if no metrics are specified.

                For predict step, it returns pred only.
        """
        # Split the batch into input and target. Target will be `None` if not provided.
        input, target = self._split_batch(batch)

        # Forward
        if self.inferer and mode in ["val", "test", "predict"]:
            pred = self.inferer(input, self)
        else:
            pred = self(input)

        pred = reshape_pred_if_single_value_prediction(pred, target)

        # Calculate the loss.
        loss = None
        if mode in ["train", "val"]:
            loss = self._calculate_loss(pred, target)

        # Apply the post-criterion activation. Necessary for measuring the metrics
        # correctly in cases when using a criterion such as `BCELossWithLogits`` which
        # requires the model to output logits, i.e. non-activated outputs.
        if self._post_criterion_activation is not None:
            pred = self._post_criterion_activation(pred)

        if mode == "predict":
            # In predict mode, skip the metrics and return the predicted value only.
            return pred
        else:
            # Calculate the metrics for the step.
            metrics = getattr(self, f"{mode}_metrics")(pred, target)
            # Log the loss and metrics for monitoring purposes only.
            self.log("loss" if mode == "train" else f"{mode}_loss", loss, on_step=True, on_epoch=True, logger=False)
            self.log_dict(metrics, on_step=True, on_epoch=True, logger=False)
            # Return the loss, metrics, input, target, and pred.
            return {"loss": loss, "metrics": metrics, "input": input, "target": target, "pred": pred}

    def _calculate_loss(
        self, pred: Union[torch.Tensor, List, Tuple, Dict], target: Union[torch.Tensor, List, Tuple, Dict, None]
    ) -> torch.Tensor:
        """_summary_

        Args:
            pred (torch.Tensor, List, Tuple, Dict, None): the predicted values from the model.
            target (torch.Tensor, List, Tuple, Dict, None): the target/label.

        Returns:
            torch.Tensor: the calculated loss.
        """
        # Keyword arguments to pass to the loss/criterion function
        kwargs = {}
        if hasarg(self.criterion.forward, "target"):
            # Add `target` argument if forward accepts it. Casting performed if specified.
            kwargs["target"] = target.to(self._cast_target_dtype_to)
        else:
            if not self._target_not_used_reported and not self.trainer.sanity_checking:
                self._target_not_used_reported = True
                logger.info(
                    f"The criterion `{get_name(self.criterion, True)}` "
                    "has no `target` argument. In such cases, the LighterSystem "
                    "passes only the predicted values to the criterion. "
                    "This is intended as a support for self-supervised "
                    "losses where target is not used. If this is not the "
                    "behavior you expected, redefine your criterion "
                    "so that it has a `target` argument."
                )

        # Type not supported.
        if not isinstance(pred, (torch.Tensor, tuple, list, dict)):
            logger.error(f"Pred type '{type(pred)}' not supported.")
            sys.exit()

        # Unpack Tuple or List. Only if num of args passed is less than or equal to num of args accepted.
        if isinstance(pred, (tuple, list)) and (len(pred) + len(kwargs)) <= countargs(self.criterion):
            return self.criterion(*pred, **kwargs)
        # Unpack Dict. Only if dict's keys match criterion's keyword arguments' names.
        elif isinstance(pred, dict) and all(hasarg(self.criterion, name) for name in pred):
            return self.criterion(**pred, **kwargs)
        # Tensor, Tuple, List, or Dict, as-is, not unpacked.
        else:
            return self.criterion(pred, **kwargs)

    def _base_dataloader(self, mode: str) -> DataLoader:
        """Instantiate the dataloader for a mode (train/val/test/predict).
        Includes a collate function that enables the DataLoader to replace
        None's (alias for corrupted examples) in the batch with valid examples.
        To make use of it, write a try-except in your Dataset that handles
        corrupted data by returning None instead.

        Args:
            mode (str): mode for which to create the dataloader ["train", "val", "test", "predict"].

        Returns:
            DataLoader: instantiated DataLoader.
        """
        dataset = getattr(self, f"{mode}_dataset")
        sampler = getattr(self, f"{mode}_sampler")
        collate_fn = getattr(self, f"{mode}_collate")

        if dataset is None:
            logger.error(f"Please specify '{mode}_dataset' in the config. Exiting")
            sys.exit()

        # Batch size is 1 when using an inference for two reasons:
        # 1) Inferer separates an input into multiple parts, forming a batch of its own.
        # 2) The val/test/pred data usually differ in their shape, so they cannot be stacked into a batch.
        batch_size = self.batch_size
        if self.inferer is not None and mode in ["val", "test", "predict"]:
            logger.info(f"Setting the '{mode}' mode dataloader to batch size of 1 because an inferer is provided.")
            batch_size = 1

        # A dataset can return None when a corrupted example occurs. This collate
        # function replaces None's with valid examples from the dataset.
        collate_fn = partial(collate_fn_replace_corrupted, dataset=dataset, default_collate_fn=collate_fn)
        return DataLoader(
            dataset,
            sampler=sampler,
            shuffle=(mode == "train" and sampler is None),
            batch_size=batch_size,
            drop_last=self.drop_last_batch,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            collate_fn=collate_fn,
        )

    def configure_optimizers(self) -> Union[Optimizer, List[Dict[str, Union[Optimizer, "Scheduler"]]]]:
        """LightningModule method. Returns optimizers and, if defined, schedulers.

        Returns:
            Optimizer or a List of Dict of paired Optimizers and Schedulers: instantiated
                optimizers and/or schedulers.
        """
        if not self.optimizers:
            logger.error("Please specify 'system.optimizers' in the config. Exiting.")
            sys.exit()
        if not self.schedulers:
            return self.optimizers

        if len(self.optimizers) != len(self.schedulers):
            logger.error("Each optimizer must have its own scheduler.")
            sys.exit()

        return [{"optimizer": opt, "lr_scheduler": sched} for opt, sched in zip(self.optimizers, self.schedulers)]

    def setup(self, stage: str) -> None:
        """Automatically called by the LightningModule after the initialization.
        `LighterSystem`'s setup checks if the required dataset is provided in the config and
        sets up LightningModule methods for the stage in which the system is.

        Args:
            stage (str): passed by PyTorch Lightning. ["fit", "validate", "test"].
        """
        # Stage-specific PyTorch Lightning methods. Defined dynamically so that the system
        # only has methods used in the stage and for which the configuration was provided.
        if not self._lightning_module_methods_defined:
            del (
                self.train_dataloader,
                self.training_step,
                self.val_dataloader,
                self.validation_step,
                self.test_dataloader,
                self.test_step,
                self.predict_dataloader,
                self.predict_step,
            )
            # `Trainer.tune()` calls the `self.setup()` method whenever it runs for a new
            #  parameter, and deleting the above methods again breaks it. This flag prevents it.
            self._lightning_module_methods_defined = True

        # Training methods.
        if stage in ["fit", "tune"]:
            self.train_dataloader = partial(self._base_dataloader, mode="train")
            self.training_step = partial(self._base_step, mode="train")

        # Validation methods. Required in 'validate' stage and optionally in 'fit' or 'tune' stage.
        if stage == "validate" or (stage in ["fit", "tune"] and self.val_dataset is not None):
            self.val_dataloader = partial(self._base_dataloader, mode="val")
            self.validation_step = partial(self._base_step, mode="val")

        # Test methods.
        if stage == "test":
            self.test_dataloader = partial(self._base_dataloader, mode="test")
            self.test_step = partial(self._base_step, mode="test")

        # Predict methods.
        if stage == "predict":
            self.predict_dataloader = partial(self._base_dataloader, mode="predict")
            self.predict_step = partial(self._base_step, mode="predict")

    def _split_batch(self, batch) -> Tuple[torch.Tensor, Optional[Any]]:
        """Split the batch into input and target. Target will be `None` if not provided.

        Args:
            batch (List, Tuple): output of the DataLoader and input to the model.

        Returns:
            Tuple(torch.Tensor, Optional[Any]): input and target.
        """
        # Check if the batch format is correct.
        if len(batch) > 2:
            raise ValueError(
                "Found more than 2 items in the batch. `LighterSystem` requires the dataloader to return either "
                "input tensor(s) only, or a two-element tuple/list consisting of input tensor(s) and target(s)."
            )

        # Report the batch split type. Only on the first call.
        if not self._batch_type_reported:
            self._batch_type_reported = True
            if len(batch) == 1:
                logger.info("Target not provided. Using `None` as target. Ignore if intended.")
            else:
                logger.info("Using the first item as input and the second item as target.")

        # Split the batch into input and target. Target will be `None` if not provided.
        input, target = (batch, None) if len(batch) == 1 else batch
        return input, target

    def _init_placeholders_for_dataloader_and_step_methods(self) -> None:
        """`LighterSystem` dynamically defines the `..._dataloader()`and `..._step()` methods
        in the `self.setup()` method. However, `LightningModule` excepts them to be defined at
        the initialization. To prevent it from throwing an error, the `..._dataloader()` and
        `..._step()` are initially defined as `lambda: None`, before `self.setup()` is called.
        """
        self.train_dataloader = self.training_step = lambda: None
        self.val_dataloader = self.validation_step = lambda: None
        self.test_dataloader = self.test_step = lambda: None
        self.predict_dataloader = self.predict_step = lambda: None



from typing import Any, Dict, Union

import sys
from datetime import datetime
from pathlib import Path

import torch
from loguru import logger
from monai.utils.module import optional_import
from pytorch_lightning import Callback, Trainer

from lighter import LighterSystem
from lighter.callbacks.utils import get_lighter_mode, is_data_type_supported, parse_data, preprocess_image
from lighter.utils.dynamic_imports import OPTIONAL_IMPORTS


class LighterLogger(Callback):
    def __init__(
        self,
        project: str,
        log_dir: str,
        tensorboard: bool = False,
        wandb: bool = False,
        input_type: str = None,
        target_type: str = None,
        pred_type: str = None,
        max_samples: int = None,
    ) -> None:
        self.project = project
        # Only used on rank 0, the dir is created in setup().
        self.log_dir = Path(log_dir) / project / datetime.now().strftime("%Y%m%d_%H%M%S")

        self.log_types = {"input": input_type, "target": target_type, "pred": pred_type}
        # Max number of samples from the batch to log.
        self.max_samples = max_samples

        self.tensorboard = tensorboard
        self.wandb = wandb

        # Running loss. Resets at each epoch. Loss is not calculated for `test` mode.
        self.loss = {"train": 0, "val": 0}
        # Epoch steps. Resets at each epoch. No `test` mode step counter as loss is not calculated for it.
        self.epoch_step_counter = {"train": 0, "val": 0}
        # Global steps. Replaces `Trainer.global_step` because it counts optimizer steps
        # instead of batch steps, which can be problematic when using gradient accumulation.
        self.global_step_counter = {"train": 0, "val": 0, "test": 0}

    def setup(self, trainer: Trainer, pl_module: LighterSystem, stage: str) -> None:
        """
        Initialize logging for the LighterSystem.
        TODO: improve this docstring.

        Args:
            trainer (Trainer): Trainer, passed automatically by PyTorch Lightning.
            pl_module (LighterSystem): LighterSystem, passed automatically by PyTorch Lightning.
            stage (str): stage of the training process. Passed automatically by PyTorch Lightning.
        """
        if trainer.logger is not None:
            logger.error("When using LighterLogger, set Trainer(logger=None).")
            sys.exit()

        if not trainer.is_global_zero:
            return

        self.log_dir.mkdir(parents=True)

        # Load the dumped config file to log it to the loggers.
        # config = yaml.safe_load(open(self.log_dir / "config.yaml"))

        # Loguru log file.
        # logger.add(sink=self.log_dir / f"{stage}.log")

        # Tensorboard initialization.
        if self.tensorboard:
            # Tensorboard is a part of PyTorch, no need to check if it is not available.
            OPTIONAL_IMPORTS["tensorboard"], _ = optional_import("torch.utils.tensorboard")
            tensorboard_dir = self.log_dir / "tensorboard"
            tensorboard_dir.mkdir()
            self.tensorboard = OPTIONAL_IMPORTS["tensorboard"].SummaryWriter(log_dir=tensorboard_dir)
            # self.tensorboard.add_hparams(config)

        # Wandb initialization.
        if self.wandb:
            OPTIONAL_IMPORTS["wandb"], wandb_available = optional_import("wandb")
            if not wandb_available:
                logger.error("Weights & Biases not installed. To install it, run `pip install wandb`. Exiting.")
                sys.exit()
            wandb_dir = self.log_dir / "wandb"
            wandb_dir.mkdir()
            self.wandb = OPTIONAL_IMPORTS["wandb"].init(project=self.project, dir=wandb_dir)
            # self.wandb.config.update(config)

    def teardown(self, trainer: Trainer, pl_module: LighterSystem, stage: str) -> None:
        if not trainer.is_global_zero:
            return
        if self.tensorboard:
            self.tensorboard.close()

    def _log_by_type(self, name: str, data: Any, log_type: str, global_step: int) -> None:
        """Logs data according to the specified type.

        Args:
            name (str): name of the data being logged.
            data (Any): data to be logged.
            log_type (str): type of the data. Supported types are 'scalar', 'image', and 'histogram'.
            global_step (int): current global step.

        Raises:
            TypeError: if the specified `log_type` is not supported.
        """
        # Log scalar
        if log_type == "scalar":
            self._log_scalar(name, data, global_step)

        # Log image
        elif log_type == "image":
            # Slice to `max_samples` only if it less than the batch size.
            if self.max_samples is not None and self.max_samples < data.shape[0]:
                data = data[: self.max_samples]
            # Preprocess a batch of images into a single, loggable, image.
            data = preprocess_image(data)
            self._log_image(name, data, global_step)

        # Log histogram
        elif log_type == "histogram":
            self._log_histogram(name, data, global_step)

        else:
            raise TypeError(f"`{log_type}` not supported for logging.")

    def _log_scalar(self, name: str, scalar: Union[int, float, torch.Tensor], global_step: int) -> None:
        """Logs the scalar.

        Args:
            name (str): name of the image to be logged.
            scalar (int, float, torch.Tensor): image to be logged.
            global_step (int): current global step.
        """
        if not isinstance(scalar, (int, float, torch.Tensor)):
            raise NotImplementedError("LighterLogger currently supports only single scalars.")
        if isinstance(scalar, torch.Tensor) and scalar.dim() > 0:
            raise NotImplementedError("LighterLogger currently supports only single scalars.")

        if isinstance(scalar, torch.Tensor):
            scalar = scalar.item()

        if self.tensorboard:
            self.tensorboard.add_scalar(name, scalar, global_step=global_step)
        if self.wandb:
            self.wandb.log({name: scalar}, step=global_step)

    def _log_image(self, name: str, image: torch.Tensor, global_step: int) -> None:
        """Logs the image.

        Args:
            name (str): name of the image to be logged.
            image (torch.Tensor): image to be logged.
            global_step (int): current global step.
        """
        if self.tensorboard:
            self.tensorboard.add_image(name, image, global_step=global_step)
        if self.wandb:
            self.wandb.log({name: OPTIONAL_IMPORTS["wandb"].Image(image)}, step=global_step)

    def _log_histogram(self, name: str, tensor: torch.Tensor, global_step: int) -> None:
        """Logs the histogram.

        Args:
            name (str): name of the image to be logged.
            tensor (torch.Tensor): tensor to be logged.
            global_step (int): current global step.
        """
        tensor = tensor.detach().cpu()

        if self.tensorboard:
            self.tensorboard.add_histogram(name, tensor, global_step=global_step)
        if self.wandb:
            self.wandb.log({name: OPTIONAL_IMPORTS["wandb"].Histogram(tensor)}, step=global_step)

    def _on_batch_end(self, outputs: Dict, trainer: Trainer) -> None:
        """Performs logging at the end of a batch/step. It logs the loss and metrics,
        and, if their logging type is specified, the input, target, and pred data.

        Args:
            outputs (Dict): output dict from the model.
            trainer (Trainer): Trainer, passed automatically by PyTorch Lightning.
        """
        if not trainer.sanity_checking:
            mode = get_lighter_mode(trainer.state.stage)
            # Accumulate the loss.
            if mode in ["train", "val"]:
                self.loss[mode] += outputs["loss"].item()
            # Logging frequency. Log only on rank 0.
            if trainer.is_global_zero and self.global_step_counter[mode] % trainer.log_every_n_steps == 0:
                # Get global step.
                global_step = self._get_global_step(trainer)

                # Log loss.
                if outputs["loss"] is not None:
                    self._log_scalar(f"{mode}/loss/step", outputs["loss"], global_step)

                # Log metrics.
                if outputs["metrics"] is not None:
                    for name, metric in outputs["metrics"].items():
                        self._log_scalar(f"{mode}/metrics/{name}/step", metric, global_step)

                # Log input, target, and pred.
                for name in ["input", "target", "pred"]:
                    if self.log_types[name] is None:
                        continue
                    # Ensure data is of a valid type.
                    if not is_data_type_supported(outputs[name]):
                        raise ValueError(
                            f"`{name}` has to be a Tensor, List[Tensor], Tuple[Tensor],  Dict[str, Tensor], "
                            f"Dict[str, List[Tensor]], or Dict[str, Tuple[Tensor]]. `{type(outputs[name])}` is not supported."
                        )
                    for identifier, item in parse_data(outputs[name]).items():
                        item_name = f"{mode}/data/{name}" if identifier is None else f"{mode}/data/{name}_{identifier}"
                        self._log_by_type(item_name, item, self.log_types[name], global_step)

            # Increment the step counters.
            self.global_step_counter[mode] += 1
            if mode in ["train", "val"]:
                self.epoch_step_counter[mode] += 1

    def _on_epoch_end(self, trainer: Trainer, pl_module: LighterSystem) -> None:
        """Performs logging at the end of an epoch. Logs the epoch number, the loss, and the metrics.
        It averages the loss and metrics over the epoch. In distributed mode, it averages over all ranks.

        Args:
            trainer (Trainer): Trainer, passed automatically by PyTorch Lightning.
            pl_module (LighterSystem): LighterSystem, passed automatically by PyTorch Lightning.
        """
        if not trainer.sanity_checking:
            mode = get_lighter_mode(trainer.state.stage)
            loss, metrics = None, None

            # Loss
            if mode in ["train", "val"]:
                # Get the accumulated loss.
                loss = self.loss[mode]
                # Reduce the loss and average it on each rank.
                loss = trainer.strategy.reduce(loss, reduce_op="mean")
                # Divide the accumulated loss by the number of steps in the epoch.
                loss /= self.epoch_step_counter[mode]

            # Metrics
            # Get the torchmetrics.
            metric_collection = getattr(pl_module, f"{mode}_metrics")
            if metric_collection is not None:
                # Compute the epoch metrics.
                metrics = metric_collection.compute()
                # Reset the metrics for the next epoch.
                metric_collection.reset()

            # Log. Only on rank 0.
            if trainer.is_global_zero:
                # Get global step.
                global_step = self._get_global_step(trainer)

                # Log epoch number.
                self._log_scalar("epoch", trainer.current_epoch, global_step)

                # Log loss.
                if loss is not None:
                    self._log_scalar(f"{mode}/loss/epoch", loss, global_step)

                # Log metrics.
                if metrics is not None:
                    for name, metric in metrics.items():
                        self._log_scalar(f"{mode}/metrics/{name}/epoch", metric, global_step)

    def _get_global_step(self, trainer: Trainer) -> int:
        """Return the global step for the current mode. Note that when Trainer
        is running Trainer.fit() and is in `val` mode, this method will return
        the global step of the `train` mode in order to correctly log the validation
        steps against training steps.

        Args:
            trainer (Trainer): Trainer, passed automatically by PyTorch Lightning.

        Returns:
            int: global step.
        """
        mode = get_lighter_mode(trainer.state.stage)
        # When validating in Trainer.fit(), return the train steps instead of the
        # val steps to correctly
        if mode == "val" and trainer.state.fn == "fit":
            return self.global_step_counter["train"]
        return self.global_step_counter[mode]

    def on_train_epoch_start(self, trainer: Trainer, pl_module: LighterSystem) -> None:
        # Reset the loss and the epoch step counter for the next epoch.
        self.loss["train"] = 0
        self.epoch_step_counter["train"] = 0

    def on_validation_epoch_start(self, trainer: Trainer, pl_module: LighterSystem) -> None:
        # Reset the loss and the epoch step counter for the next epoch.
        self.loss["val"] = 0
        self.epoch_step_counter["val"] = 0

    def on_train_batch_end(self, trainer: Trainer, pl_module: LighterSystem, outputs: Any, batch: Any, batch_idx: int) -> None:
        self._on_batch_end(outputs, trainer)

    def on_validation_batch_end(
        self, trainer: Trainer, pl_module: LighterSystem, outputs: Any, batch: Any, batch_idx: int, dataloader_idx: int
    ) -> None:
        self._on_batch_end(outputs, trainer)

    def on_test_batch_end(
        self, trainer: Trainer, pl_module: LighterSystem, outputs: Any, batch: Any, batch_idx: int, dataloader_idx: int
    ) -> None:
        self._on_batch_end(outputs, trainer)

    def on_train_epoch_end(self, trainer: Trainer, pl_module: LighterSystem) -> None:
        self._on_epoch_end(trainer, pl_module)

    def on_validation_epoch_end(self, trainer: Trainer, pl_module: LighterSystem) -> None:
        self._on_epoch_end(trainer, pl_module)

    def on_test_epoch_end(self, trainer: Trainer, pl_module: LighterSystem) -> None:
        self._on_epoch_end(trainer, pl_module)
